package org.zeroturnaround.javarebel.integration.spring;

import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Iterator;
import java.util.List;
import org.springframework.beans.factory.xml.ParserContext;
import org.springframework.context.annotation.ComponentScanBeanDefinitionParser;
import org.w3c.dom.Element;
import org.zeroturnaround.javarebel.Logger;
import org.zeroturnaround.javarebel.LoggerFactory;
import org.zeroturnaround.javarebel.StopWatch;

/**
 * Requires Spring 2.1 M3 or later.
 * <p>
 * Should only be used if class
 * org.springframework.context.annotation.ComponentScanBeanDefinitionParser
 * exists.
 */
public class AnnotationBeanDefinitionManager {
  
  private static final Logger log = LoggerFactory.getInstance();
  
  private static final List componentScanners = 
    Collections.synchronizedList(new ArrayList());
  private static Method reparseMethod;

  private static volatile long lastScan = 0;
  
  private static final long SCAN_TIMEOUT = 500; // ms
  private static final Object SCAN_LOCK = new Object();
  private static final AtomicBoolean scanInProgress = new AtomicBoolean(false);

  static {
    try {
      reparseMethod = ComponentScanBeanDefinitionParser.class.getDeclaredMethod(
          "reparse", new Class[] {
              Element.class,
              ParserContext.class});
    } catch (SecurityException e) {
      log.errorEcho(e);
    } catch (NoSuchMethodException e) {
      log.errorEcho(e);
    }
  }

  private static class ComponentScanner {
    private ComponentScanBeanDefinitionParser componentScanner; 
    private Element element;
    private ParserContext parserContext;

    public ComponentScanner(ComponentScanBeanDefinitionParser componentScanner,
        Element element, ParserContext parserContext) {
      this.componentScanner = componentScanner;
      this.element = element;
      this.parserContext = parserContext;
    }

    public void scan() throws Throwable {
      try {
        reparseMethod.invoke(componentScanner, new Object[] {element, parserContext});
        // XXX Causes VisitEstonia to fail due to early instantiation!
//        if (parserContext.getRegistry() instanceof DefaultListableBeanFactory) {
//          // XXX Is this safe? See {@link DefaultListableBeanFactoryCBP}
//          ((DefaultListableBeanFactory) parserContext.getRegistry()).preInstantiateSingletons();
//        }
      } catch (IllegalAccessException e) {
        log.errorEcho(e);
        throw e;
      } catch (InvocationTargetException e) {
        log.errorEcho(e.getTargetException());
        throw e.getTargetException();
      }
    }
  }

  public static void registerComponentScanner(
      ComponentScanBeanDefinitionParser componentScanner, 
      Element element, 
      ParserContext parserContext) {
    componentScanners.add(new ComponentScanner(componentScanner, element, parserContext));
  }

  public static void checkRescan() {
    if (lastScan + 1000 > System.currentTimeMillis())
      return;       

    rescan();

    lastScan = System.currentTimeMillis();
  }
  
  private static void rescan() {
    String threadName = null;
    if (log.isEnabled())
      threadName = Thread.currentThread().getName();
    
    if (scanInProgress.compareAndSet(false, true)) {
      // Starting to scan
      if (log.isEnabled())
        log.log("JRebel-Spring: Starting to scan annotations <" + threadName + ">");
      
      try {
        doRescan();
      }
      finally {
        // Finished scanning
        scanInProgress.set(false);
        if (log.isEnabled())
          log.log("JRebel-Spring: Finished scanning annotations <" + threadName + ">");
        
        // Notify other threads
        synchronized (SCAN_LOCK) {
          SCAN_LOCK.notifyAll();
        }
      }
    }
    else {
      // Wait for the scan to be complete (with a timeout)
      if (log.isEnabled())
        log.log("JRebel-Spring: Waiting for the annotations scanner <" + threadName + ">");
      
      while (scanInProgress.get()) {
        synchronized (SCAN_LOCK) {
          try {
            SCAN_LOCK.wait(SCAN_TIMEOUT);
          } catch (InterruptedException e) {
          }
        }
      }

      if (log.isEnabled())
        log.log("JRebel-Spring: Finished waiting for the annotations scanner <" + threadName + ">");
    }
  }

  private static void doRescan() {
    StopWatch sw = log.createStopWatch("AnnotationBeanDefinitionManager.rescan");

    try {
      for (Iterator i = new ArrayList(componentScanners).iterator(); i.hasNext();) {
        ComponentScanner cs = (ComponentScanner) i.next();
        cs.scan();
      }      
    } 
    catch (Throwable e) {
      if (e instanceof Error) throw (Error) e;
      log.error(e);
    }
    finally {
      if (sw != null) sw.stop();
    }
  }
}
