package org.zeroturnaround.javarebel.integration.spring.prop;

import java.lang.reflect.Field;
import java.util.Enumeration;
import java.util.Properties;
import org.springframework.beans.BeansException;
import org.springframework.beans.factory.BeanInitializationException;
import org.springframework.beans.factory.config.BeanDefinition;
import org.springframework.beans.factory.config.PropertyOverrideConfigurer;
import org.springframework.beans.factory.config.PropertyResourceConfigurer;
import org.zeroturnaround.javarebel.LoggerFactory;

/**
 * @author Rein Raudjärv
 * 
 * @see PropertyOverrideConfigurer
 */
public class PropertyOverrideConfigurerWrapper extends PropertyResourceConfigurerWrapper {
  
  private static final Field BEAN_NAME_SEPARATOR_FIELD;
  private static final Field IGNORE_INVALID_KEYS_FIELD;
  
  static {
    BEAN_NAME_SEPARATOR_FIELD = getDeclaredField(
        PropertyOverrideConfigurer.class,
        "beanNameSeparator");
    
    IGNORE_INVALID_KEYS_FIELD = getDeclaredField(
        PropertyOverrideConfigurer.class,
        "ignoreInvalidKeys");
    
    if (!initialized)
      log.log("JRebel-Spring: Could not initialize " + PropertyOverrideConfigurerWrapper.class.getName());
  }
  
  private String getBeanNameSeparator() {
    try {
      return (String) BEAN_NAME_SEPARATOR_FIELD.get(getConfigurer());
    } catch (RuntimeException e) {
      throw e;
    } catch (Exception e) {
      throw new RuntimeException(e);
    }
  }  
  
  private boolean isIgnoreInvalidKeys() {
    try {
      return ((Boolean) IGNORE_INVALID_KEYS_FIELD.get(getConfigurer())).booleanValue();
    } catch (RuntimeException e) {
      throw e;
    } catch (Exception e) {
      throw new RuntimeException(e);
    }
  }  

  public PropertyOverrideConfigurerWrapper(PropertyResourceConfigurer configurer) {
    super(configurer);
  }
  
  /**
   * Processes the given bean definition by overriding property values
   * using the given properties.
   * 
   * @param beanName bean name.
   * @param bd bean definition.
   * @param props properties used.
   */
  protected void processProperties(String beanName, BeanDefinition bd, Properties props) {
    if (log.isEnabled())
      log.log("JRebel-Spring: Processing override properties for bean '" + beanName + "'");
    
    for (Enumeration names = props.propertyNames(); names.hasMoreElements();) {
      String key = (String) names.nextElement();
      try {
          processKey(beanName, bd, key, props.getProperty(key));
      }
      catch (BeansException ex) {
          String msg = "Could not process key '" + key + "' in PropertyOverrideConfigurer";
          if (!isIgnoreInvalidKeys()) {
              throw new BeanInitializationException(msg, ex);
          }
          LoggerFactory.getInstance().error(ex);
      }
    }
  }
  
  private void processKey(String beanName, BeanDefinition bd, String key, String value) {
    int separatorIndex = key.indexOf(getBeanNameSeparator());
    if (separatorIndex == -1) {
        throw new BeanInitializationException("Invalid key '" + key +
                "': expected 'beanName" + getBeanNameSeparator() + "property'");
    }
    String name = key.substring(0, separatorIndex);
    String property = key.substring(separatorIndex+1);
    
    if (beanName.equals(name)) {
      if (log.isTraceEnabled())
        log.trace("JRebel-Spring: Seeting bean '" + beanName + "' property '" + property + "': " + value);
      
      bd.getPropertyValues().addPropertyValue(property, value);
    }
  }

}
