package org.zeroturnaround.javarebel.integration.spring.prop;

import java.lang.reflect.Field;
import java.lang.reflect.Method;
import java.util.HashSet;
import java.util.Properties;
import java.util.Set;
import org.springframework.beans.BeansException;
import org.springframework.beans.factory.BeanDefinitionStoreException;
import org.springframework.beans.factory.config.BeanDefinition;
import org.springframework.beans.factory.config.BeanDefinitionVisitor;
import org.springframework.beans.factory.config.PropertyPlaceholderConfigurer;
import org.springframework.beans.factory.config.PropertyResourceConfigurer;
import org.springframework.util.StringValueResolver;

/**
 * @author Rein Raudjärv
 * 
 * @see PropertyPlaceholderConfigurer
 */
public class PropertyPlaceholderConfigurerWrapper extends PropertyResourceConfigurerWrapper {

  private static final Method PARSE_STRING_VALUE_METHOD;
  private static final Field NULLVALUE_FIELD;

  static {
    PARSE_STRING_VALUE_METHOD = getDeclaredMethod(
        PropertyPlaceholderConfigurer.class,
        "parseStringValue",
        new Class[] { String.class, Properties.class, Set.class });

    NULLVALUE_FIELD = getDeclaredField(
        PropertyPlaceholderConfigurer.class,
    "nullValue");

    if (!initialized)
      log.log("JRebel-Spring: Could not initialize " + PropertyPlaceholderConfigurerWrapper.class.getName());
  }

  private String parseStringValue(String strVal, Properties props, Set visitedPlaceholders) {
    try {
      return (String) PARSE_STRING_VALUE_METHOD.invoke(getConfigurer(), new Object[] { strVal, props, visitedPlaceholders });
    } catch (RuntimeException e) {
      throw e;
    } catch (Exception e) {
      throw new RuntimeException(e);
    }
  }

  private String getNullValue() {
    try {
      return (String) NULLVALUE_FIELD.get(getConfigurer());
    } catch (RuntimeException e) {
      throw e;
    } catch (Exception e) {
      throw new RuntimeException(e);
    }
  }  

  public PropertyPlaceholderConfigurerWrapper(PropertyResourceConfigurer configurer) {
    super(configurer);
  }

  /**
   * Processes the given bean definition by resolving property placeholder values
   * using the given properties.
   * 
   * @param beanName bean name.
   * @param bd bean definition.
   * @param props properties used.
   */
  protected void processProperties(String beanName, BeanDefinition bd, Properties props) {
    if (log.isEnabled())
      log.log("JRebel-Spring: Processing placeholder properties for bean '" + beanName + "'");

    StringValueResolver valueResolver = new PlaceholderResolvingStringValueResolver(props);
    BeanDefinitionVisitor visitor = new BeanDefinitionVisitor(valueResolver);

    try {
      visitor.visitBeanDefinition(bd);
    }
    catch (BeanDefinitionStoreException ex) {
      throw new BeanDefinitionStoreException(bd.getResourceDescription(), beanName, ex.getMessage());
    }
  }

  /**
   * BeanDefinitionVisitor that resolves placeholders in String values,
   * delegating to the <code>parseStringValue</code> method of the
   * containing class.
   */
  private class PlaceholderResolvingStringValueResolver implements StringValueResolver {

    private final Properties props;

    public PlaceholderResolvingStringValueResolver(Properties props) {
      this.props = props;
    }

    public String resolveStringValue(String strVal) throws BeansException {
      String value = parseStringValue(strVal, this.props, new HashSet());
      value = value.equals(getNullValue()) ? null : value;

      if (log.isTraceEnabled())
        log.trace("JRebel-Spring: Resoleved " + strVal + " -> " + value);

      return value;
    }
  }

}
