package org.zeroturnaround.javarebel.integration.spring.prop;

import java.util.Iterator;
import java.util.Map;
import java.util.WeakHashMap;
import org.springframework.beans.BeansException;
import org.springframework.beans.factory.config.BeanDefinition;
import org.springframework.beans.factory.config.ConfigurableListableBeanFactory;
import org.springframework.beans.factory.config.PropertyOverrideConfigurer;
import org.springframework.beans.factory.config.PropertyPlaceholderConfigurer;
import org.springframework.beans.factory.config.PropertyResourceConfigurer;
import org.zeroturnaround.javarebel.Logger;
import org.zeroturnaround.javarebel.LoggerFactory;

/**
 * @author Rein Raudjärv
 * 
 * @see PropertyPlaceholderConfigurer
 * @see PropertyOverrideConfigurer
 * @see PropertyPlaceholderConfigurerWrapper
 * @see PropertyOverrideConfigurerWrapper
 */
public class PropertyResourceConfigurerHelper {
  
  private static final Logger log = LoggerFactory.getInstance();
  
  // Map<ConfigurableListableBeanFactory, PropertyResourceConfigurerHelper> helper per each bean factory.
  private static final Map helpers = new WeakHashMap();
  
  public static synchronized PropertyResourceConfigurerHelper get(ConfigurableListableBeanFactory beanFactory) {
    PropertyResourceConfigurerHelper helper = (PropertyResourceConfigurerHelper) helpers.get(beanFactory);
    if (helper == null) {
      helper = new PropertyResourceConfigurerHelper();
      helpers.put(beanFactory, helper);
    }
    return helper;
  }
  
  // Map<PropertyPlaceholderConfigurer, PropertyPlaceholderConfigurerWrapper>
  private final Map placeholders = new WeakHashMap();
  
  // Map<PropertyOverrideConfigurer, PropertyOverrideConfigurerWrapper>
  private final Map overrides = new WeakHashMap();
  
  public void register(PropertyResourceConfigurer configurer) {
    if (configurer instanceof PropertyPlaceholderConfigurer)
      register((PropertyPlaceholderConfigurer) configurer); 
    else if (configurer instanceof PropertyOverrideConfigurer)
      register((PropertyOverrideConfigurer) configurer); 
  }
  public void register(PropertyPlaceholderConfigurer configurer) {
    try {
      placeholders.put(configurer, new PropertyPlaceholderConfigurerWrapper(configurer));
      log.log("JRebel-Spring: Registered " + configurer);
    } catch (Exception e) {
      log.log("JRebel-Spring: Could not register PropertyPlaceholderConfigurer " + configurer);
      log.error(e);
    }
  }
  public void register(PropertyOverrideConfigurer configurer) {
    try {
      overrides.put(configurer, new PropertyOverrideConfigurerWrapper(configurer));
      log.log("JRebel-Spring: Registered " + configurer);
    } catch (Exception e) {
      log.log("JRebel-Spring: Could not register PropertyPlaceholderConfigurer " + configurer);
      log.error(e);
    }
  }
  
  public void apply(String beanName, BeanDefinition bd) throws BeansException {
    try {
      applyPlaceholderConfigurers(beanName, bd);
      applyOverrideConfigurers(beanName, bd);
    } catch (Exception e) {
      log.log("JRebel-Spring: Could not process properties of bean '" + beanName + "'");
      log.error(e);
    }
  }
  
  private void applyPlaceholderConfigurers(String beanName, BeanDefinition bd) throws BeansException {
    for (Iterator it = placeholders.values().iterator(); it.hasNext();) {
      PropertyPlaceholderConfigurerWrapper wrapper = (PropertyPlaceholderConfigurerWrapper) it.next();
      wrapper.processProperties(beanName, bd);
    }
  }
  
  private void applyOverrideConfigurers(String beanName, BeanDefinition bd) throws BeansException {
    for (Iterator it = overrides.values().iterator(); it.hasNext();) {
      PropertyOverrideConfigurerWrapper wrapper = (PropertyOverrideConfigurerWrapper) it.next();
      wrapper.processProperties(beanName, bd);
    }
  }
  
}
