package org.zeroturnaround.javarebel.integration.spring.prop;

import java.io.IOException;
import java.lang.ref.WeakReference;
import java.lang.reflect.Field;
import java.lang.reflect.Method;
import java.util.Properties;
import org.springframework.beans.factory.config.BeanDefinition;
import org.springframework.beans.factory.config.PropertyResourceConfigurer;
import org.springframework.core.io.Resource;
import org.springframework.core.io.support.PropertiesLoaderSupport;
import org.zeroturnaround.javarebel.Logger;
import org.zeroturnaround.javarebel.LoggerFactory;

/**
 * @author Rein Raudjärv
 * 
 * @see PropertiesLoaderSupport
 * @see PropertyResourceConfigurer
 */
public abstract class PropertyResourceConfigurerWrapper {
  
  private static final long CHECK_INTERVAL = 2000L; // 2 seconds
  
  protected static final Logger log = LoggerFactory.getInstance();
  
  private static final Method MERGE_PROPERTIES_METHOD;
  private static final Method CONVERT_PROPERTIES_METHOD;
  private static final Field LOCATIONS_FIELD;
  protected static boolean initialized = true;
  
  private final WeakReference configurer;
  
  private long previousCheck;
  private long previousModified;
  
  private Properties properties;
  
  static {
    MERGE_PROPERTIES_METHOD = getDeclaredMethod(
        PropertiesLoaderSupport.class,
        "mergeProperties",
        new Class[0]);
    
    CONVERT_PROPERTIES_METHOD = getDeclaredMethod(
        PropertyResourceConfigurer.class,
        "convertProperties",
        new Class[] { Properties.class });
    
    LOCATIONS_FIELD = getDeclaredField(
        PropertiesLoaderSupport.class,
        "locations");
    
    if (!initialized)
      log.log("JRebel-Spring: Could not initialize " + PropertyResourceConfigurerWrapper.class.getName());
  }
  
  protected static Method getDeclaredMethod(Class clazz, String methodName, Class[] paramTypes) {
    Method m = null;
    try {
      m = clazz.getDeclaredMethod(methodName, paramTypes);
      m.setAccessible(true);
    } catch (Exception e) {
      LoggerFactory.getInstance().error(e);
      initialized = false;
    }
    return m;
  }
  
  protected static Field getDeclaredField(Class clazz, String fieldName) {
    Field f = null;
    try {
      f = clazz.getDeclaredField(fieldName);
      f.setAccessible(true);
    } catch (Exception e) {
      LoggerFactory.getInstance().error(e);
      initialized = false;
    }
    return f;
  }
  
  protected Properties mergeProperties() {
    try {
      return (Properties) MERGE_PROPERTIES_METHOD.invoke(getConfigurer(), new Object[0]);
    } catch (RuntimeException e) {
      throw e;
    } catch (Exception e) {
      throw new RuntimeException(e);
    }
  }
  
  protected void convertProperties(Properties props) {
    try {
      CONVERT_PROPERTIES_METHOD.invoke(getConfigurer(), new Object[] { props });
    } catch (RuntimeException e) {
      throw e;
    } catch (Exception e) {
      throw new RuntimeException(e);
    }
  }
  
  protected Resource[] getLocations() {
    try {
      return (Resource[]) LOCATIONS_FIELD.get(getConfigurer());
    } catch (RuntimeException e) {
      throw e;
    } catch (Exception e) {
      throw new RuntimeException(e);
    }
  }
  
  public PropertyResourceConfigurerWrapper(PropertyResourceConfigurer configurer) {
    this.configurer = new WeakReference(configurer);
    this.previousCheck = System.currentTimeMillis();
    this.previousModified = lastModified();
    this.properties = loadProperties();
  }
  
  protected PropertyResourceConfigurer getConfigurer() {
    return (PropertyResourceConfigurer) configurer.get();
  }
  
  /**
   * @return latest last modified date of the resources.
   */
  private long lastModified() {
    long result = 0L;
    
    Resource[] locations = getLocations();
    if (log.isTraceEnabled())
      log.trace("JRebel-Spring: Locations of " + getConfigurer() + " are " + locations);
    
    if (locations != null) {
      for (int i = 0; i < locations.length; i++) {
        Resource location = locations[i];
        long lm = 0L;
        try {
          lm = location.lastModified();
        } catch (IOException e) {
          log.error(e);
        }
        if (log.isTraceEnabled())
          log.trace("JRebel-Spring: Last modified of " + location + " is " + lm);
        
        if (lm != 0L && lm > result)
          result = lm;
      }
    }
    
    return result;
  }
  
  /**
   * @return <code>true</code> if any of the resources have been updated.
   */
  private boolean isModified() {
    long newModified = lastModified();
    if (newModified == 0)
      return false;
    
    if (newModified != previousModified) {
      previousModified = newModified;
      return true;
    }
    
    return false;
  }
  
  /**
   * Skips the is modified check if last check was less than second ago.
   */
  private boolean checkModified() {
    long now = System.currentTimeMillis();
    long delta = now - previousCheck;
    previousCheck = now;
    if (delta < CHECK_INTERVAL) {
      if (log.isTraceEnabled())
        log.trace("JRebel-Spring: Checking modified too often: " + getConfigurer());      
      return false;
    }
    
    return isModified();
  }

  /**
   * Returns properties using cache if resources are up to date.
   */
  protected synchronized Properties getProperties() {
    if (checkModified())
      properties = loadProperties();
    else if (log.isTraceEnabled())
      log.trace("JRebel-Spring: Returning properties from cache: " + getConfigurer());
    
    return properties;
  }
  
  /**
   * Loads properties.
   * 
   * @return properties loaded.
   */
  private Properties loadProperties() {
    if (log.isEnabled())
      log.log("JRebel-Spring: Loading properties: " + getConfigurer());

    Properties mergedProps = mergeProperties();

    // Convert the merged properties, if necessary.
    convertProperties(mergedProps);

    return mergedProps;
  }
  
  /**
   * Processes the given bean definition by applying the properties.
   * 
   * @param beanName bean name.
   * @param bd bean definition.
   */
  public void processProperties(String beanName, BeanDefinition bd) {
    if (!initialized)
      return;
    
    Properties props = getProperties();
    processProperties(beanName, bd, props);
  }
  
  protected abstract void processProperties(String beanName, BeanDefinition bd, Properties props);

}
